class SendPackageRequest
  class WrongNumberOfElements < StandardError; end

  include ActionView::Helpers::TextHelper

  attr_accessor :sender, :raw_xml, :xml_doc, :error_condition, :error_message
  attr_accessor :to_names, :from_name, :subject, :body, :source_name, :source_paths
  attr_accessor :valid, :well_formed, :package_valid, :from_valid, :recipients_valid

  def initialize(sender, request_xml)
    self.sender = sender
    self.raw_xml = request_xml
    parse_request
  end

  def parse_request
    self.valid =
        validate_sender_permission &&
        parse_xml &&
        validate_package &&
        validate_from &&
        validate_recipients
  end

  def validate_sender_permission
    if sender.uploads_allowed?
      true
    else
      self.error_condition = 'unauthorized'
      self.error_message = I18n.t('app.models.send_package_request.validate_sender_error')
      false
    end
  end

  def parse_xml
    raw_xml =~ /^<\?xml.*\?>(.*)/m # workaround for wrong encoding in Outlook Add-In
    self.xml_doc = XmlParser.parse_xml(raw_xml, true)
    self.well_formed = true
   rescue XmlParser::InvalidXML => e
     self.error_condition = 'xml-not-well-formed'
     self.error_message = I18n.t('app.models.send_package_request.parse_xml_error')
     self.well_formed = false
  end

  def validate_package
    package = get_elements(xml_doc, 'package', 1, 1).first
    self.to_names = get_elements(package, 'to', 1).map(&:content)
    self.from_name = get_elements(package, 'from', 0, 1).first.try(:content)
    self.subject = get_elements(package, 'subject', 1, 1).first.content
    self.body = get_elements(package, 'body', 0, 1).first.try(:content)

    if sources = get_elements(package, 'sources', 0, 1).first
      source = get_elements(sources, 'source', 1, 1).first
      paths = get_elements(source, 'paths', 1, 1).first
      self.source_name = get_elements(source, 'name', 1, 1).first.content
      self.source_paths = get_elements(paths, 'path', 1).map(&:content)
    end

    self.package_valid = true
  rescue WrongNumberOfElements => e
    self.error_condition = 'package-not-well-formed'
    self.error_message = e.message
    self.package_valid = false
  end

  def validate_from
    if from_name.nil? ||
        [
          sender.name.downcase, sender.email.downcase,
          "#{sender.name.downcase}@#{EConfiguration.first.e_server_address.downcase}",
        ].include?(from_name.downcase)
      self.from_valid = true
    else
      self.error_condition = 'sender-authentication-mismatch'
      self.error_message = I18n.t('app.models.send_package_request.from_name_error')
      self.from_valid = false
    end
  end

  def validate_recipients
    valid_names = []
    invalid_names = []
    to_names.each do |name|
      if cname = canonical_name(name)
        valid_names << cname
      else
        invalid_names << name
      end
    end
    if !EConfiguration.first.ignore_invalid_recipients? && invalid_names.any?
      self.error_condition = 'invalid-recipient'
      self.error_message =
        "#{pluralize(invalid_names.count, I18n.t('app.models.send_package_request.invalid_recipient_error'))}: " +
          invalid_names.join(', ')
      self.recipients_valid = false
    else
      self.to_names = valid_names
      self.recipients_valid = true
    end
  end

  def canonical_name(name)
    case
    when sender.valid_recipient_name?(name)
      name
    when name.downcase =~ /(.*)@#{EConfiguration.first.e_server_address.downcase}$/ &&
        sender.valid_recipient_name?($1)
      $1
    else
      nil
    end
  end

  # Search under this node for between min and max number of matching elements.
  # WrongNumberOfElements is raised if element count doesn't fall between min
  # and max (min and max are ignored if nil).
  def get_elements(node, path, min = nil, max = nil)
    elements = node.css(path)
    if min && elements.count < min
      raise WrongNumberOfElements,
        I18n.t('app.models.send_package_request.too_few_elements', {path: path, min: pluralize(min, 'time')})
    elsif max && elements.count > max
      raise WrongNumberOfElements,
        I18n.t('app.models.send_package_request.too_many_elements', {path: path, max: pluralize(max, 'time')})
    else
      elements
    end
  end

  def send_params(options = {})
    {
      title: subject,
      note: (body || ""),
      sender: sender,
      recipients: to_names,
      sources: [{ id: 1 }],
      skip_path_validation: options[:skip_path_validation],
      skip_metadata_validation: options[:skip_metadata_validation],
      skip_recipient_validation: options[:skip_recipient_validation],
      delete_after_download_policy: sender.effective_delete_after_download_policy(EPackage::DO_NOTHING)
    }
  end

  def formatted_recipient_list
    to_names.join("\n")
  end

  def formatted_path_list
    source_paths.try(:join, "\n")
  end

  def source_share_id
    source_share.try(:id)
  end

  def source_share
    SourceShare.find_by_name(source_name)
  end

  def valid?; valid; end
  def well_formed?; well_formed; end
  def package_valid?; package_valid; end
  def from_valid?; from_valid; end
  def recipients_valid?; recipients_valid; end
end
