require_dependency 'utils'
require_dependency 'portable_contacts_searcher'
require_dependency 'transfer_settings_selector'

class User < ActiveRecord::Base
  extend RailsFriendlyObservable
  include MultiServerHooks
  include ActionController::UrlWriter # for generating path to delegate DS authentication to master
  include TransferSettingsSelector

  concerned_with :portable_contacts
  concerned_with :inbox_service
  concerned_with :privileges
  concerned_with "permissions/workgroup_permissions"
  concerned_with "permissions/package_permissions"
  concerned_with "permissions/user_permissions"
  concerned_with "permissions/self_permissions"
  concerned_with "permissions/invitations_permissions"


  # package after-download expiration override policies
  POLICY_ACCEPT_SERVER_DEFAULT   = 0
  POLICY_OVERRIDE_SERVER_DEFAULT = 1

  OPTIONALLY_REQUIRED_FIELDS = [:account_expires, :password_expires,
                                :allowed_ip_addresses_for_login,
                                :allowed_ip_addresses_for_download,
                                :allowed_ip_addresses_for_upload]

  OPTIONALLY_REQUIRED_FIELDS.each do |optionally_required_param|
    if optionally_required_param != :password_expires
      validates_presence_of(optionally_required_param, :if => Proc.new{ |u| !u.external_email_user? }) if RbConfig.optionally_require_user_field?(optionally_required_param)
    else
      validates_presence_of(optionally_required_param, :if => Proc.new{ |u| u.is_a?(LocalUser) }) if RbConfig.optionally_require_user_field?(optionally_required_param)
    end
  end

  has_one :registration
  has_many :ad_groups, :through => :user_ad_groups
  has_many :user_ad_groups, :dependent => :destroy
  has_many :e_memberships, :dependent => :destroy
  has_many :e_workgroups, :through => :e_memberships
  has_many :readable_e_workgroups,
           :through => :e_memberships,
           :source => :e_workgroup,
           :conditions => ['e_memberships.can_read = ?', true]
  has_many :writable_e_workgroups,
           :through => :e_memberships,
           :source => :e_workgroup,
           :conditions => ['e_memberships.can_write = ?', true]
  has_many :administered_e_workgroups,
           :through => :e_memberships,
           :source => :e_workgroup,
           :conditions => ['e_memberships.is_wg_admin = ?', true]
  has_many :distribution_lists, foreign_key: :user_id, dependent: :destroy

  has_many :user_deliveries, as: :recipient

  has_many :received_invitations, :foreign_key => :invitee_id, :dependent => :destroy, class_name: 'Invitation'

  has_many :share_memberships, as: :member, dependent: :destroy
  has_many :source_shares, through: :share_memberships, class_name: 'SourceShare', source: :share,
      conditions: {source_access: Share::LIMITED_ACCESS}

  has_many :user_histories, dependent: :destroy
  has_many :user_profile_data, dependent: :destroy, validate: false
  has_many :user_profile_fields, through: :user_profile_data

  has_many :oauth_tokens, dependent: :destroy

  has_many :previous_passwords, :order => 'created_at DESC', :dependent => :destroy

  belongs_to :home_server
  belongs_to :authorization_domain

  validates_length_of :description, :allowed_ip_addresses_for_upload, :allowed_ip_addresses_for_download, :allowed_ip_addresses_for_login, :maximum => 255, :allow_blank => true

  validates_associated :user_profile_data, if: :validate_user_profile_data?

  attr_accessible :notify_on_receive, :notify_on_workgroup_package_if_sender, :notify_on_download, :notify_downloader_on_download, :rows_per_page,
      :suppress_connect_install_dialog, :authorization_domain, :user_profile_data_attributes, :user_profile_data, :override_default_password_policy,
      :enforce_password_history, :password_history_limit, :allow_receipt_cc_edit, :disable_connect, :notify_on_validation_failure_if_sender,
      :notify_on_validation_failure_if_recipient, :override_security_sender_quota, :sender_quota_limit, :locale, :show_only_completed_packages


  before_validation :sanitize_and_strip_tags

  before_save :recalculate_home_server
  after_save :remove_internal_user_from_contacts
  after_save :send_account_created_email, :if => :send_account_created_email?

  before_destroy :hold_pending_deliveries

  named_scope :admins, :conditions => {:admin => true}
  named_scope :non_admin, :conditions => ["admin = ?", false]
  named_scope :super_admins, :conditions => ['admin=? AND privileges LIKE ?', true, '%super%']
  named_scope :logged_in, lambda { {:conditions => ['session_expiry_time > ?', Time.now]} }

  accepts_nested_attributes_for :user_profile_data

  validates_numericality_of :sender_quota_limit, only_integer: true, allow_nil: true, greater_than_or_equal_to: 0, less_than: 999999

  def self.sync_ad_users
    AuthorizationDomain.active.each { |ad| ad.sync_ad_users }
  end

  def self.deactivate_expired_users
    logger.info("User.deactivate_expired_users...")
    expired_users = AuthorizableUser.all(conditions: [
      'account_activated = ? AND account_expires = ? AND expiration_date <= ?',
      true, true, Date.today
    ])
    expired_users.each do |u|
      Utils.log_debug("User account #{u.id} expired. Deactivating...")
      u.deactivate
    end
  end

  def self.deactivate_or_remove_inactive_users
    logger.info("User.deactivate_or_remove_inactive_users...")

    e_config = EConfiguration.first

    User.non_admin.each do |user|
      if user.eligible_for_inactivity_removal?(e_config)
        logger.warn("User account #{user.id} inactive. Removing...")
        user.destroy
      elsif user.eligible_for_inactivity_deactivation?(e_config)
        logger.warn("User account #{user.id} inactive. Deactivating...")
        user.deactivate
      end
    end
  end

  def hold_pending_deliveries
    if self.id == 1
      logger.warn("User with id 1 cannot be destroyed, stacktrace #{caller}")
      return false
    end
    deliveries = Delivery.find(:all, :conditions => ["sender_id = ? AND release_policy = ? AND state = ?", self.id, Delivery::RELEASE_POLICY_ON_DATE, Delivery::STATE_HELD])
    deliveries.each do |d|
      d.update_non_released_from_hash(self, {:release_policy => Delivery::RELEASE_POLICY_NEVER, :release_date => nil})
    end
  end

  def self.no_common_workgroup(user1, user2)
    (user1.e_workgroups & user2.e_workgroups).empty?
  end

  DISPLAY_FORMAT_USERNAME   = 'username'
  DISPLAY_FORMAT_FULL_NAME  = 'full_name'

  def self.display_formats
    [['Username', DISPLAY_FORMAT_USERNAME],
     ['Full Name', DISPLAY_FORMAT_FULL_NAME]]
  end

  def display_name
    case EConfiguration.first.account_display_name_format
      when DISPLAY_FORMAT_FULL_NAME
        full_name
      else
        name
    end
  end

  def add_failed_login_attempt(remote_ip)
    _arr = self.failed_login_attempts.strip.split("\n") rescue [] # text column can't have default value
    #if this failed login attempt was more than 5 minutes ago, then delete
    _arr.delete_if { |t| Time.now - Time.parse(t) > EConfiguration.user_deactivation_login_interval * 60 }
    if _arr.length>=EConfiguration.user_deactivation_login_count-1
      logger.warn("SECURITY: Failed login attempt by '#{self.id}' at [#{Time.now.to_s}]. Deactivating user.")
      self.lock unless is_a?(DirectoryServiceUser)
      if (MultiServer::Synchronizer.slave_server? && !self.external_email_user?)
         self.needs_sync_to_master = true
         self.save!
       end
      self.clear_failed_login_attempts
    else
      logger.warn("SECURITY: Failed login attempt by '#{self.id}' at [#{Time.now.to_s}]")
      self.failed_login_attempts = _arr.join("\n") + "\n#{Time.now.to_s}"
      self.save
    end
  end

  # if import succeeds, return params hash for creating user
  def self.import_ldap_user(authorization_domain, ldap_user, ad_group = nil, attributes = {})
    ds_user = new_ds_user(authorization_domain, ldap_user)

    ldap_attrs = ldap_user.attributes
    if ldap_attrs.nil?
      raise AuthorizationDomain::MissingUser, "user with distinguished name '#{user_dn.to_s}' not found on DS server"
    else
      user_attrs = get_ad_user_params(ldap_attrs, ad_group)
      attributes.merge!(user_attrs[:user])
      set_ad_user_attributes(ds_user, attributes)
      if ds_user.save
        # protected attributes
        [:role, :can_send_to_external_email, :can_send_normal_packages, :can_send_to_all_faspex_users].each do |attr|
          ds_user.update_attribute(attr, attributes[attr]) if !attributes[attr].nil?
        end
        user_attrs[:created_user_id] = ds_user.id
        user_attrs[:aduser] = ds_user
        user_attrs[:created_user] = true
        return user_attrs
      else
        Utils.log_warn("Import DS user: Error saving user: #{ds_user.errors.full_messages.join("\n")}")
        raise RuntimeError, ds_user.errors.full_messages.join("\n")
      end
    end
  end

  def self.new_ds_user(authorization_domain, ldap_user)
    external_user = ExternalUser.find_by_name(ldap_user.name) || ExternalUser.find_by_name(ldap_user.email)
    if external_user
      external_user.type = DirectoryServiceUser.name
      external_user.becomes(DirectoryServiceUser)
      external_user.name = ldap_user.name
      external_user.authorization_domain = authorization_domain
      external_user.inherit_group_settings if external_user.ad_groups.any?
      external_user.save!

      # Convert recipient name is user deliveries in case username has changed or external flag is present
      UserDelivery.find_all_by_recipient_id(external_user.id).map{ |d| d.update_attribute(:recipient_name, external_user.name)}

      DirectoryServiceUser.find_by_name(ldap_user.name)
    else
      DirectoryServiceUser.new(name: ldap_user.name, authorization_domain: authorization_domain)
    end
  end

  def self.get_ad_user_params(user_hash, ad_group)
    hash = {}
    user_hash.merge!({
      :password               => 'x' * 40,
      :ad_individual_import   => ad_group.nil?,
    })
    user_hash.delete(:account_activated) unless RbConfig.ds_sync_active_state?

    if ad_group
      user_hash.merge!({
        :expiration_date                   => ad_group.expiration_date,
        :account_expires                   => ad_group.account_expires,
        :uploads_allowed                   => ad_group.uploads_allowed,
        :downloads_allowed                 => ad_group.downloads_allowed,
        :forwarding_allowed                => ad_group.forwarding_allowed,
        :can_create_from_remote            => ad_group.can_create_from_remote,
        :can_send_to_external_email        => ad_group.can_send_to_external_email,
        :can_send_normal_packages          => ad_group.can_send_normal_packages,
        :can_send_to_all_faspex_users      => ad_group.can_send_to_all_faspex_users,
        :allowed_ip_addresses_for_upload   => ad_group.allowed_ip_addresses_for_upload,
        :allowed_ip_addresses_for_download => ad_group.allowed_ip_addresses_for_download,
        :allowed_ip_addresses_for_login    => ad_group.allowed_ip_addresses_for_login,
        # to prevent license count from being exceeded, user will be activated by UsersController#import_user
        :account_activated         => false
      })

      if ad_group.override_transfer_settings?
        hash[:override_transfer_settings] = true
        user_hash.merge!({
          :kbps_initial_target_upload_rate   => ad_group.kbps_initial_target_upload_rate.to_s,
          :kbps_initial_target_download_rate => ad_group.kbps_initial_target_download_rate.to_s,
          :lock_minrate_policy               => ad_group.lock_minrate_policy,
          :kbps_max_target_upload_rate       => ad_group.kbps_max_target_upload_rate.to_s,
          :kbps_max_target_download_rate     => ad_group.kbps_max_target_download_rate.to_s
        })
      end
    end

    hash[:user] = user_hash
    hash
  end

  #############################################################################
  # INSTANCE METHODS
  #############################################################################

  # user is local to this server
  def local?
    # home_server_id = 0 means this server
    home_server_id == 0 || home_server == HomeServer.this_server
  end

  def allowed_to_send_normal_packages?
    can_send_normal_packages == User::Permissionable::YES ||
        can_send_normal_packages == User::Permissionable::DEFAULT && EConfiguration.first.users_can_send_normal_packages?
  end

  def allowed_to_send_to_all_faspex_users?
    can_send_to_all_faspex_users == User::Permissionable::YES ||
        can_send_to_all_faspex_users == User::Permissionable::DEFAULT && EConfiguration.first.users_can_send_to_all_faspex_users?
  end

  def allowed_to_specify_delete_after_download?
    allow_user_specified_delete_after_download == User::Permissionable::YES ||
        allow_user_specified_delete_after_download == User::Permissionable::DEFAULT && EConfiguration.first.allow_user_specified_delete_after_download?
  end

  def effective_delete_after_download_policy(package_level_policy)
    if allowed_to_specify_delete_after_download?
      package_level_policy
    elsif override_server_delete_after_download?
      delete_after_download_policy
    else
      EConfiguration.first.delete_after_download_policy
    end
  end

  def home_server_id
    if super.nil?
      recalculate_home_server!
      update_attribute(:home_server_id, self.home_server_id) unless new_record?
    end
    super
  end

  alias :rails_defined_home_server :home_server
  def home_server
    if home_server_id == 0 # 0 indicates this server
      HomeServer.this_server
    else
      rails_defined_home_server
    end
  end

  def home_servers
    [home_server]
  end

  # recalculates the home server unless the home server setting is manually overridden
  def recalculate_home_server(rules=nil)
    home_server_id # This triggers a recalculation of the home server if it has not been set yet
    if HomeServer.this_server.master? && !override_home_server?
      recalculate_home_server!(rules)
    end
  end

  # forcibly recalculate the home server, regardless of manual override
  def recalculate_home_server!(rules=nil)
    matching_hs = home_server_by_matching_rule(rules)
    if matching_hs
      self.home_server_id = matching_hs.id
    else
      self.home_server_id = external_email_user? ? 0 : HomeServer.default_server.id
    end
  end

  def home_server_by_matching_rule(rules=nil)
    rules = HomeServerRule.find(:all, :order => :position) if rules.nil?
    rules.each do |rule|
      return rule.home_server if self.match_rule?(rule)
    end
    nil # no rules matched
  end

  def calculated_home_server(rules = nil)
    home_server_by_matching_rule(rules) || HomeServer.default_server
  end

  def match_rule?(rule)
    #if rule applies to only external email recipients, then only check email address regex
    if rule.external_email_user?
      #get rid of non-external users
      return false unless self.is_a?(ExternalUser)
      #now looking at only external users, check if they match the email regex, if it exists
      if rule.email_pattern.blank?
        return true
      else
        return self.email =~ Regexp.new(rule.email_pattern)
      end
    end

    #non-external users
    return false if self.is_a?(ExternalUser)
    unless rule.user_name_pattern.blank?
      return false if self.name !~ Regexp.new(rule.user_name_pattern)
    end
    unless rule.email_pattern.blank?
      return false if self.email !~ Regexp.new(rule.email_pattern)
    end

    #if non-AD user, then pass if there are no AD-related fields, fail if there are
    return (rule.ad_dn_pattern.blank? &&
            rule.ad_group_pattern.blank? &&
            rule.ldap_field_pattern1.blank? &&
            rule.ldap_field_pattern2.blank? &&
            rule.ldap_field_pattern3.blank?) unless self.is_a?(DirectoryServiceUser)

    #AD-users left, filter through the rest of the fields
    unless rule.ad_group_pattern.blank?
      _groups = self.ad_groups.reject{ |g| g.name !~ Regexp.new(rule.ad_group_pattern) }
      return false if _groups.empty?
    end
    unless rule.ad_dn_pattern.blank?
      attribute = self.get_ldap_attribute(:dn)
      return false if attribute.nil? || attribute !~ Regexp.new(rule.ad_dn_pattern)
    end
    unless rule.ldap_field1.blank? || rule.ldap_field_pattern1.blank?
      attribute = self.get_ldap_attribute(rule.ldap_field1)
      return false if attribute.nil? || attribute !~ Regexp.new(rule.ldap_field_pattern1)
    end
    unless rule.ldap_field2.blank? || rule.ldap_field_pattern2.blank?
      return false unless self.is_a?(DirectoryServiceUser)
      attribute = self.get_ldap_attribute(rule.ldap_field2)
      return false if attribute.nil? || attribute !~ Regexp.new(rule.ldap_field_pattern2)
    end
    unless rule.ldap_field3.blank? || rule.ldap_field_pattern3.blank?
      return false unless self.is_a?(DirectoryServiceUser)
      attribute = self.get_ldap_attribute(rule.ldap_field3)
      return false if attribute.nil? || attribute !~ Regexp.new(rule.ldap_field_pattern3)
    end

    #AD user, passed all filters
    true
  end

  def get_ldap_attribute(key)
    YAML.load(self.ldap_attributes).to_hash[key.to_sym]
  end

  def dn
    get_ldap_attribute('distinguishedname')
  end

  def active?
    self.account_activated == true && !saml_config_disabled?
  end

  def saml_configuration
    SamlConfiguration.find_by_id(saml_configuration_id)
  end

  def saml_config_disabled?
    saml_configuration.present? && !saml_configuration.enabled?
  end

  def inactive?
    !active? || saml_config_disabled?
  end

  def expires?
    self.account_expires
  end

  def expired?
    self.account_expires && self.expiration_date && (self.expiration_date <= Date.today)
  end

  def clear_failed_login_attempts
    self.failed_login_attempts = ''
    self.save
  end

  def lock
    self.locked = true
    if save && EConfiguration.first.allow_users_to_unlock?
      send_reset_password_email
    end
  end

  # Must keep track of whether this account was deactivated manually (Faspex) or automatically (DS import)
  def deactivate(manual=true)
    if self.is_a?(ExternalUser)
      Utils.log_debug("Error deactivating user: Attempting to deactivate external user")
      return false, "Error deactivating user: Attempting to deactivate external user"
    end

    if saml_config_disabled?
      return false, "Error deactivating user: SAML configuration is currently disabled"
    end

    self.manually_deactivated = manual
    self.account_activated = false
    if self.save
      hold_pending_deliveries
      return true, nil
    else
      puts "Unable to save"
      Utils.log_warn("Error deactivating user: #{self.errors.full_messages.join("\n")}")
      return false, "Error deactivating user: #{self.errors.full_messages.join("\n")}"
    end
  end

  def activate
    if self.is_a?(ExternalUser)
      Utils.log_debug("Error activating user '#{id}': Attempting to activate external user")
      return false, "Error activating user '#{id}': Attempting to activate external user"
    end

    if saml_config_disabled?
      return false, "Error activating user: SAML configuration is currently disabled"
    end

    self.account_activated = true
    self.locked = false
    self.last_activity_at = Time.now
    self.manually_deactivated = false
    if self.save
      return true, nil
    else
      Utils.log_warn("Error activating user '#{name}': #{self.errors.full_messages.join("\n")}")
      return false, "Error activating user '#{name}': #{self.errors.full_messages.join("\n")}"
    end
  end

  def account_activated=(val)
    # when setting account_activated = true, we also want to make sure account is unlocked
    if val
      self.locked = false
    end
    super(val)
  end

  def workgroup_deliveries
    WorkgroupDelivery.find_all_by_recipient_id(self.e_workgroups.map(&:id))
  end

  def e_workgroups_through_ad
    _w = []
    self.user_ad_groups.each do |uag|
      _w |= uag.ad_group.e_workgroups
    end
    _w
  end

  def member_of_workgroup_via_other_adgroups?(workgroup_id, adgroup_id)
    member_via_another_adgroup = false
    self.ad_groups.each do |other_adgroup|
      next if (other_adgroup.id == adgroup_id)
      other_adgroup.e_workgroups.each do |wg|
        if wg.id == workgroup_id
          member_via_another_adgroup = true
          break
        end
      end
      break if member_via_another_adgroup
    end
    member_via_another_adgroup
  end

  def set_active(is_active)
    if is_active && !self.active?
      success, msg = self.activate
    elsif !is_active && self.active?
      success, msg = self.deactivate
    else
      return true
    end

    if success
      return true
    else
      return false, msg
    end
  end

  def external_email_user?
    #self.is_a?(ExternalUser)
    self.type == ExternalUser.name || self.is_a?(ExternalUser)
  end

  def self.external_user_ear(passcode)
    passcode = passcode.to_s
    return ExternalUser.all.select{|element| element.received_invitations.first.present? && element.received_invitations.first.passcode == passcode}[0].external_user_ear rescue '0'
  end

  def authorizable_user?
    #[DirectoryServiceUser, LocalUser].any? { |klass| self.is_a?(klass) }
    [DirectoryServiceUser.name, LocalUser.name, SelfRegisteredUser.name].any? { |klass_name| self.type == klass_name }
  end

  def update_login_info(host_ip, remote_ip)
    #store user's current IP address info
    self.last_login_time = Time.now.strftime("%Y-%m-%d %H:%M:%S")
    self.last_server_addr = host_ip
    self.last_remote_addr = remote_ip
    self.failed_login_attempts = ''

    # Unless concurrent login is prevented, use the same login uuid for all sessions
    if EConfiguration.prevent_concurrent_login? || self.login_uuid.nil?
      self.login_uuid = SecureRandom.uuid
    end

    # if this a slave server, mark record for sync to master
    self.needs_sync_to_master = true if HomeServer.this_server.slave?

    unless self.save
      logger.error('Error trying to save changes to user record while logging in: ' +
                   self.errors.full_messages.join("\n"))
    end
  end

  def send_reset_password_email
    token = ResetPasswordToken.generate(self).token
    BackgroundNotifier.deliver_reset_password_notification(self, token)
  end

  def portable_contacts(params)
    PortableContactsSearcher.new(self, params).search
  end

  def addressbook(page = 0, current_recipients = "", criteria = "", options = {})

    recipients, pagination_info =
      query_user_domain(criteria.to_s, 15, page, "detailedSummary",
                        nil, 'preferredUsername', options[:filter_type])

    addressbook = []

    recipients.each do |r|
      if /(.+)\(DUP\)/.match(r["preferredUsername"])
        preferredUsername = $1.strip
      else
        preferredUsername = r["preferredUsername"]
      end
      rhash = {:addressable => preferredUsername}
      case r["tags"]
      when "users"
        name     = r["preferredUsername"]
        informal = r["emails"]

        unless informal == name
          informal = "(" + informal + ")"
          informal += " " + r["displayName"]
        else  # login == email address
          informal = r["displayName"]
        end

        rhash[:address_book_entry] =
          "<strong>#{ERB::Util.h name}</strong> " +
          "#{ERB::Util.h informal}"
        rhash[:current_recipient] = current_recipients.include?(r["preferredUsername"].strip.downcase)
      when "external-users"
        rhash[:address_book_entry] =
          "<strong>#{ERB::Util.h preferredUsername}</strong> (external)"
        rhash[:current_recipient] = current_recipients.include?(preferredUsername.strip.downcase)
        rhash[:addressable] += ' (external)'
      when "workgroups"
        rhash[:addressable] = "*#{ERB::Util.h r["preferredUsername"]}"
        rhash[:address_book_entry] =
          "* <strong>#{ERB::Util.h r["preferredUsername"]}</strong>  (workgroup)</span>"
        rhash[:current_recipient] = current_recipients.include?("*" + r["preferredUsername"].to_s.strip.downcase)
      when "distribution-lists"
        rhash[:addressable] = "#{ERB::Util.h r["preferredUsername"]}"
        rhash[:address_book_entry] =
            "<strong>#{ERB::Util.h r["preferredUsername"]}</strong>  (distribution list)</span>"
        rhash[:current_recipient] = current_recipients.include?(r["preferredUsername"].strip.downcase)
      end
      addressbook << rhash
    end

    return addressbook, pagination_info
  end

  def valid_user_recipients
    # what users can I send to?
    _users = []
    if allowed_to_send_to_all_faspex_users?
      _users += AuthorizableUser.active
    end
    _users += valid_wg_member_recipients
    _users << self
    _users.uniq
  end

  def valid_wg_member_recipients
    recipients = e_workgroups.inject(Set.new) do |users, wg|
      case
        when admin_of?(wg) || wg.wg_member_access == EWorkgroup::MUTUAL_ACCESS
          users + wg.members
        when wg.wg_member_access = EWorkgroup::WG_ADMIN_ACCESS
          users + wg.e_memberships.wg_admins.collect(&:user)
      end
    end
    recipients.to_a
  end

  def valid_wg_recipients
    # what workgroups can I send to?
    _workgroups = []
    EWorkgroup.find(:all).each do |workgroup|
      if workgroup.can_user_send_to_wg?(self)
        _workgroups << workgroup
      end
    end
    _workgroups.uniq
  end

  def valid_recipients_hash
    @valid_recipients_hash ||= begin
      recipients = valid_user_recipients + valid_wg_recipients
      {}.tap do |recipient_hash|
        recipients.each {|r| recipient_hash[r.name] = r}
      end
    end
  end

  def valid_recipients
    valid_recipients_hash.values
  end

  def lookup_recipient(name, include_external=true, for_ccs=false)
    name = name.sub(/^\*/, '')
    lookup_distribution_list_recipient(name) ||
        lookup_workgroup_recipient(name) ||
        lookup_user_recipient(name) ||
        (include_external && lookup_external_recipient(name, for_ccs))
  end

  # Personal distribution list takes precedence
  def lookup_distribution_list_recipient(name)
    recipient = distribution_lists.find_by_name(name)

    if recipient.nil? && can_see_global_distribution_lists?
      recipient = DistributionList.global.find_by_name(name)
    end

    recipient
  end

  def lookup_workgroup_recipient(name)
    # need to check workgroups that user is not a member of (for open workgroups)
    wg = EWorkgroup.find_by_e_wg_name(name)
    if wg && wg.can_user_send_to_wg?(self)
      wg
    else
      nil
    end
  end

  def lookup_user_recipient(name)
    user = AuthorizableUser.find_by_name(name)

    if user.nil? && !ExternalUser.find_by_name(name)
      user = AuthorizableUser.find_by_email(name)
    end

    if user && can_send_to_user?(user)
      user
    else
      nil
    end
  end
  
  def lookup_external_recipient(name, for_ccs=false)
    # Strip external flag
    if /(.+)\(external\)/.match(name)
      name = $1.strip
    end

    faspex_user_as_external = EConfiguration.allow_external_send_to_faspex_user? ? true : !AuthorizableUser.exists?(email: name)
    if EConfiguration.send_to_external_email_allowed? &&
        can_send_to_external_email? &&
        Utils.valid_email?(name) &&
        faspex_user_as_external
      add_external_contact(name)
      ExternalUser.find_or_create_by_name(name) rescue external_user_email(name)
    elsif for_ccs && Utils.valid_email?(name)
      add_external_contact(name)
      ExternalUser.find_or_create_by_name(name) rescue external_user_email(name)
    else
      nil
    end
  end

  def external_user_email name
    existing = ExternalUser.find_by_email(name)
    return existing if existing.present?

    ExternalUser.create(:name => "#{name} (DUP)", :email => name)
  end

  def valid_recipient?(recipient)
    lookup_recipient(recipient.name).present?
  end

  def valid_recipient_name?(name)
    lookup_recipient(name).present?
  end

  def query_user_domain(criteria, count, page, filter_by = nil, fields = nil, sort_by = 'preferredUsername', filter_type = nil)
    criteria = criteria[1..-1] if (criteria[0,1] == '*')

    # NOTE: this is a hack that uses the portable contacts code paths to perform a
    #       search on the users. The filtering code should be refactored so that
    #       this method and the portable contacts methods can use common code
    #       without being coupled to each other
    results, pagination_info, fields_to_return, format =
      portable_contacts(
        :pc_path1    => '@me',
        :pc_path2    => '@all',
        :count       => count,
        :startIndex  => page.to_i * count.to_i,
        :filterBy    => filter_by,
        :filterValue => criteria,
        :filterOp    => 'contains',
        :fields      => fields,
        :sortBy      => sort_by,
        :filterType  => filter_type
      )

    # put workgroups in a separate array to be
    # shown first
    # @todo - et - sorting contacts needs a big overhaul
    workgroups     = []
    non_workgroups = []
    results.each do |r|
      if r["tags"] == "workgroups"
        workgroups << r
      else
        non_workgroups << r
      end
    end

    return workgroups + non_workgroups, pagination_info
  end

  def match_recipients(criteria, options = {})
    results, pagination_info = query_user_domain(criteria,
                                                 15,
                                                 0,
                                                 'detailedSummary',
                                                 nil,
                                                 'preferredUsername',
                                                 options[:filter_type])
    recipients = []
    results.each do |r|
      if /(.+)\(DUP\)/.match(r["preferredUsername"])
        name = $1.strip
      else
        name = r["preferredUsername"]
      end

      case r["tags"]
      when "workgroups"
        name = "*" + name
        informal = "(workgroup)"
      when "users"
        informal = r["emails"]
        unless informal == name
          informal = "(" + informal + ")"
          informal += " " + r["displayName"]
        else  # login == email address
          informal = r["displayName"]
        end
      when "external-users"
        name += ' (external)'
      when "distribution-lists"
        informal = "(distribution list)"
      end

      recipients << {
        name:     ERB::Util.h(name),  # Escape early since this will be rendered as json
        informal: ERB::Util.h(informal)
      }
    end
    recipients
  end

  def can_send_to_user?(other_user)
    if other_user == self
      return true
    elsif !other_user.active?
      return false
    elsif other_user.is_a?(SelfRegisteredUser) && !other_user.approved_at
      return false
    elsif (allowed_to_send_to_all_faspex_users?) && other_user.authorizable_user?
      return true
    elsif other_user.is_a?(ExternalUser)
      return false unless can_send_to_external_email?
      delivery_to_other_user = UserDelivery.first(
        conditions: ["recipient_id = ? AND deliveries.sender_id = ?", other_user.id, self.id],
        joins: :delivery
      )
      return true if (not delivery_to_other_user.nil?)
    elsif other_user.invitations.map(&:invitee).include?(self)
      return true
    else
      return true if self.can_send_to_user_because_of_group_memberships?(other_user)
    end
    return false
  end

  def can_send_to_user_because_of_group_memberships?(other_user)
    e_workgroups.each do |workgroup|
      if workgroup.has_user?(other_user)
        return true if admin_of?(workgroup)

        case workgroup.wg_member_access
        when EWorkgroup::MUTUAL_ACCESS
          return true
        when EWorkgroup::WG_ADMIN_ACCESS
          return true if other_user.admin_of?(workgroup)
        end
      end
    end
    false
  end

  #--------------------------------------------------------------------------------------------------------------------------
  # User Service attribute related methods
  #--------------------------------------------------------------------------------------------------------------------------

  def is_wg_admin?(workgroup)
    membership = get_membership(workgroup)
    return membership.is_wg_admin if membership
    false
  end

  def is_wg_admin_for_any_wg?
    !EMembership.first(:conditions => ["user_id = ? AND is_wg_admin = '1'", self.id]).nil?
  end

  def get_membership_id(workgroup)
    get_membership(workgroup).id
  end

  def get_membership(workgroup)
    EMembership.find(:first, :conditions => { :e_workgroup_id => workgroup.id, :user_id => id })
  end

  def download_from_ip_allowed?(ip)
    ip_matches_list?(ip, allowed_ip_addresses_for_download)
  end

  def upload_from_ip_allowed?(ip)
    ip_matches_list?(ip, allowed_ip_addresses_for_upload)
  end

  def login_from_ip_allowed?(ip)
    ip_matches_list?(ip, allowed_ip_addresses_for_login)
  end

  def ip_matches_list?(ip, ip_address_list)
    return true if ip_address_list.blank?
    arr = ip_address_list.split(',')
    arr.each do |allowed_ip|
      return true if ip.starts_with?(allowed_ip.strip.split('*').first)
    end
    false
  end

  def ip_matches_last_remote_addr?(ip)
    match = ip.split('*').first
    last_remote_addr.starts_with?(match)
  end

  def update_protected_attributes(params)
    params = HashWithIndifferentAccess.new(params)[:user]
    self.admin = params[:admin] if !params[:admin].nil?
    self.can_send_to_external_email = params[:can_send_to_external_email] if !params[:can_send_to_external_email].nil?
    self.can_send_normal_packages = params[:can_send_normal_packages] if !params[:can_send_normal_packages].nil?
    self.can_send_to_all_faspex_users = params[:can_send_to_all_faspex_users] if !params[:can_send_to_all_faspex_users].nil?
    self.save
  end

  def add_external_contact(external_email)
    if Contact.first(:conditions => ["user_id = ? AND email = ?", self.id, external_email]).nil?
      Contact.create(:user_id => self.id, :email => external_email)
    end
  end

  #^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
  # End of User Service attribute related methods
  #^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

  def full_email_address_descriptor
    if external_email_user?
      return '"' + self.email.to_s + '" <' + self.email.to_s + '>'
    end
    '"' + self.first_name.to_s + " " + self.last_name.to_s + '" <' + self.email.to_s + '>'
  end
  
  def email_address_descriptor
    '"' + self.email.to_s + '" <' + self.email.to_s + '>'
  end

  def full_email_cc_list_descriptor
    if (_cc_list=self.received_notification_cc).nil?
      ''
    else
      _cc_list.split(/[,; \n]/).
        collect{ |a| a.strip }.
        reject{ |a| a.blank? }.
        collect{ |e| "\"#{e}\" <#{e}>"}.
        join(', ')
    end
  end

  def full_email_cc_list
    if (_cc_list=self.received_notification_cc).nil?
      []
    else
      _cc_list.split(/[,; \n]/).
          collect{ |a| a.strip }.
          reject{ |a| a.blank? }
    end
  end

  def full_name_untruncated
    [ first_name, last_name ].map(&:presence).compact.join(' ').presence || self.name || ''
  end

  def full_name
    name = [ first_name, last_name ].map(&:presence).compact.join(' ').presence || self.name || ''

    name = "#{name[0..30]}..." if name.length > 30

    name
  end

  # implement callback (called before save)
  def validate
    if !self.received_notification_cc.blank?
      self.received_notification_cc.
        split(/[,; \n]/).collect(&:strip).reject(&:blank?).
        each{ |email_address|
        unless Utils.valid_email?(email_address)
          errors.add(:received_notification_cc, "#{email_address} is an invalid email address")
        end
      }
    end

    ip_regex = /\A(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\z/
    if !self.allowed_ip_addresses_for_upload.blank?
      ul_add_arr = self.allowed_ip_addresses_for_upload.split(",").collect! {|i| i.strip}
      ul_add_arr.each do |ul_add|
        unless ul_add =~ ip_regex || is_valid_ip_address_range?(ul_add)
          errors.add(:allowed_ip_addresses_for_upload, "#{ul_add} is an invalid IP address")
        end
      end
    end
    if !self.allowed_ip_addresses_for_download.blank?
      dl_add_arr = self.allowed_ip_addresses_for_download.split(",").collect! {|i| i.strip}
      dl_add_arr.each do |dl_add|
        unless dl_add =~ ip_regex || is_valid_ip_address_range?(dl_add)
          errors.add(:allowed_ip_addresses_for_download, "#{dl_add} is an invalid IP address")
        end
      end
    end
    if !self.allowed_ip_addresses_for_login.blank?
      login_add_arr = self.allowed_ip_addresses_for_login.split(",").collect! {|i| i.strip}
      login_add_arr.each do |login_add|
        unless login_add =~ ip_regex || is_valid_ip_address_range?(login_add)
          errors.add(:allowed_ip_addresses_for_login, "#{login_add} is an invalid IP address")
        end
      end
    end

    if account_activated? && expired? && MultiServer::Synchronizer.master_server? &&
      !kind_of?(TemplateUser) # prevent validations on Template User
                              # TODO - modularize these validations
      errors.add(:expiration_date, " has already passed")
    end
  end

  def is_valid_ip_address_range?(ip)
    return true if ip=='*'
    return false if ip.last!='*' || ip.split("*").size > 1
    ip_arr = ip.split("*")[0].split(".")
    return false if ip_arr.size > 4 || ip_arr.size < 1
    ip_arr.each do |i|
      return false unless i=~/\A\d{1,3}\z/ && i.to_i <= 255 && i.to_i >= 0 # !=~ num_regex
    end
    true
  end

  def workgroup_admin?(wg_id)
    EMembership.find_by_e_workgroup_id_and_user_id(wg_id, self.id).is_wg_admin rescue false
  end

  # Sets the object guid for this user from a string which is interpreted as raw bytes (this is only relevant for DirectoryServiceUsers)
  def set_ad_objectguid_from_bytes(bytes)
    self.ad_objectguid = Utils.encode_bytes_as_hex(bytes)
  end

  # Returns a representation of the object guid for this user as raw bytes (this is only relevant for DirectoryServiceUsers)
  def get_ad_objectguid_as_bytes
    Utils.decode_bytes_from_hex(self.ad_objectguid)
  end

  def record_history(category, msg)
    UserHistory.create(:user_id => id, :category => category, :msg => msg)
  end

  def pretty_type
    case self
    when LocalUser            then "Faspex"
    when DirectoryServiceUser then self.saml? ? "SAML (#{ERB::Util.h saml_configuration.name})" : "DS"
    when ExternalUser         then "External"
    end
  end

  def delete_after_download_override_policy
    return POLICY_OVERRIDE_SERVER_DEFAULT if self.override_server_delete_after_download
    return POLICY_ACCEPT_SERVER_DEFAULT
  end

  def delete_after_download_policy
    EPackage.delete_after_download_policy(delete_packages_after_download?, delete_packages_after_download_condition)
  end

  def delete_after_download_policy=(policy)
    policy = policy.to_i
    self.delete_packages_after_download = EPackage.delete_after_download?(policy)
    self.delete_packages_after_download_condition = policy
  end

  def journal_instance?
    true
  end

  def self.changed_fields_to_ignore
    ["last_login_time", "updated_at"]
  end

  def owns_as_wg_admin(other_user)
    return false if !is_wg_admin_for_any_wg?
    return (other_user.creator_name == name)
  end

  def addressable_name
    name
  end

  def role
    return "admin"   if admin_with_privilege?(PERM_SUPER)
    return "manager" if admin?
    return "user"
  end

  def role=(newrole)
    case newrole
    when "admin"
      self.admin = true
      self.grant_privilege(PERM_SUPER)
    when "manager"
      self.admin = true
      self.revoke_privilege(PERM_SUPER)
    when "user"
      self.admin = false
      self.revoke_privilege(PERM_SUPER)
    end
  end

  def super_admin?
    admin_with_privilege?(PERM_SUPER)
  end

  # takes an array of whitelisted attributes and assigns them values given in the attribute_hash
  def set_whitelisted_attributes(whitelist, attribute_hash)
    attribute_hash.each_pair do |key, value|
      if whitelist.include?(key.to_sym)
        self[key] = value unless value.nil?
      end
    end
    set_special_whitelisted_attributes(whitelist, HashWithIndifferentAccess.new(attribute_hash))
  end

  def update_whitelisted_attributes(whitelist, attribute_hash)
    set_whitelisted_attributes(whitelist, attribute_hash)
    self.save
  end

  def home_server_name=(name)
    home_server_id_will_change!
    self.home_server = HomeServer.find_by_name(name)
  end

  def status
    case
    when saml_config_disabled? then I18n.t('app.models.user.status.saml_disabled')
    when pending? then I18n.t('app.models.user.status.pending')
    when locked?  then I18n.t('app.models.user.status.locked')
    when active?  then I18n.t('app.models.user.status.active')
    when !active? then I18n.t('app.models.user.status.inactive')
    end
  end

  def pending?
    registration.present?
  end

  def available_workgroups
    EWorkgroup.all - self.e_workgroups
  end
  alias_method :workgroups, :e_workgroups

  def add_to_workgroup(workgroup)
    EMembership.find_or_create_by_user_id_and_e_workgroup_id(self.id, workgroup.id)
  end

  def hide_user_dir?
    case hide_user_dir
    when DatabaseConstants::DEFAULT
      EConfiguration.first.hide_user_dir_default?
    when DatabaseConstants::YES
      true
    when DatabaseConstants::NO
      false
    else
      raise 'Invalid value for hide_user_dir'
    end
  end

  def can_see_global_distribution_lists?
    case can_see_global_distribution_lists
    when DatabaseConstants::DEFAULT
      EConfiguration.first.can_see_global_distribution_lists_default?
    when DatabaseConstants::YES
      true
    else
      false
    end
  end

  def global_distribution_lists
    if can_see_global_distribution_lists?
      DistributionList.global
    else
      []
    end
  end

  def multi_server_record_identifier
    name
  end

  def original_multi_server_record_identifier
    name_was
  end

  # needs to be in this class because multiserver sync creates objects of type
  # User, not SelfRegisteredUser
  def send_account_created_email
    token = ResetPasswordToken.generate(self, true).token
    BackgroundNotifier.deliver_account_approved self, token
  end

  def send_account_created_email?
    # convert to_i to avoid milliseconds throwing things off
    approved_at.to_i != approved_at_was.to_i && home_server == HomeServer.this_server
  end

  def suppress_connect_install_dialog?
    !!FASPEX_CONFIG['AlwaysSuppressConnectInstallDialog'] || super
  end

  # check if this server is the user's home server
  # admins can log in anywhere
  # managers can only log in to slave if it is their home server
  def can_login_to_this_server?
    local? || super_admin? || (admin? && HomeServer.this_server.master?)
  end

  def cannot_login_to_this_server?
    !can_login_to_this_server?
  end

  def accessible_shares
    # union of my shares and my directory groups' shares
    SourceShare.public | source_shares | ad_groups.map(&:source_shares).flatten.uniq
  end

  def force_local_login?
    false
  end

  def eligible_for_inactivity_removal?(e_config)
      if self.saml?
        e_config.remove_saml_users_on_inactivity &&
            (last_active_at < e_config.saml_user_removal_inactivity_interval.to_i.days.ago) &&
            registration.nil? # not a pending self-registered user
      elsif self.is_a?(DirectoryServiceUser)
        e_config.remove_ds_users_on_inactivity &&
            (last_active_at < e_config.ds_user_removal_inactivity_interval.to_i.days.ago) &&
            registration.nil? # not a pending self-registered user
      else
        e_config.remove_users_on_inactivity &&
            (last_active_at < e_config.user_removal_inactivity_interval.to_i.days.ago) &&
            registration.nil? # not a pending self-registered user
      end
  end

  def eligible_for_inactivity_deactivation?(e_config)
    active? &&
        e_config.deactivate_users_on_inactivity && (last_active_at < e_config.user_deactivation_inactivity_interval.to_i.days.ago) &&
        registration.nil? # not a pending self-registered user
  end

  def last_active_at
    last_activity_at || last_login_time || created_at
  end

  def record_activity
    if new_record?
      self.last_activity_at = Time.now
    else
      touch :last_activity_at
    end
  end

  def saml?
    self.is_a?(DirectoryServiceUser) && authorization_domain_id.nil?
  end

  def ldap?
    self.is_a?(DirectoryServiceUser) && !authorization_domain_id.nil?
  end

  # SAML config specific for SAML users, global otherwise
  def profile_fields
    saml? ? saml_configuration.custom_fields : UserProfileField.enabled_fields
  end

  def build_user_profile_data(user_profile_data_attributes_params=nil)
    #Following if block is needed for cases where while a user is being created/updated an admin adds new or deletes existing user profile fields.
    #In such a case we have to re-build the user_profile_data only for the newly added user_profile_field.
    #So that repeated calls (in case of validation errors) to create/update do not keep "building" redundant user_profile_data we check against the params
    #hash before building new user_profile_data
    if user_profile_data_attributes_params
      (profile_fields.map(&:id) - user_profile_data_attributes_params.map {|k, v| v["user_profile_field_id"].to_i}).each do |id|
        user_profile_data.build(user_profile_field: UserProfileField.find(id)) unless user_profile_fields.map(&:id).include? id
      end
    else
      profile_fields.each do |field|
        user_profile_data.build(user_profile_field: field) unless user_profile_fields.include?(field)
      end
    end

    # to keep the ordering of the user profile fields consistent on the UI
    user_profile_data.sort_by! { |u| u.user_profile_field_id }
  end

  def save_with_user_profile_data_validation(force=false)
    @validate_user_profile_data = self.is_a?(DirectoryServiceUser) ? false : true
    @validate_user_profile_data = true if force
    save
  end

  def validate_user_profile_data?
    (defined? @validate_user_profile_data) && @validate_user_profile_data
  end

  def user_hash
    out = {}
    out[:id] = id.to_s
    out[:name] = name
    out[:first_name] = first_name
    out[:last_name] = last_name
    out[:email] = email
    out[:account_activated] = account_activated?
    out[:account_expires] = account_expires?
    out[:admin] = admin?
    out[:downloads_allowed] = downloads_allowed?
    out[:uploads_allowed] = uploads_allowed?
    out[:can_send_normal_packages] = can_send_normal_packages?
    out[:can_send_to_all_faspex_users] = can_send_to_all_faspex_users?
    out[:can_send_to_external_email] = can_send_to_external_email?
    out[:created_at] = created_at.iso8601 if created_at.present?
    out[:updated_at] = updated_at.iso8601 if updated_at.present?
    out[:delete_packages_after_download] = delete_packages_after_download?

    out[:user_profile_data_attributes] = user_profile_data.map {|upd| upd.to_api_hash }

    out[:description] = description unless description.blank?

    out
  end

  def memberships_hash
    out = {}
    out[:memberships] = e_memberships.map{ |wgm| wgm.to_hash }

    out
  end

  def to_user_api_hash
    user_hash.merge(memberships_hash)
  end

  def configuration_hash
    out = {}
    out[:id] = id
    out[:user_hash] = Digest::MD5.hexdigest(user_hash.to_s)

    if MetadataProfile.default.present?
      out[:metadata_profile_id] = MetadataProfile.default.id
      out[:metadata_hash] = Digest::MD5.hexdigest(MetadataProfile.default.to_hash.to_s)
    end

    out[:memberships] = e_memberships.map{ |wgm|
      m = {
          id: wgm.id,
          e_workgroup_id: wgm.e_workgroup_id,
          e_workgroup_type: wgm.e_workgroup.type_name,
          membership_hash: Digest::MD5.hexdigest(wgm.to_hash.to_s),
      }

      if wgm.e_workgroup.metadata_profile.present?
        m[:metadata_profile_id] = wgm.e_workgroup.metadata_profile.id
        m[:metadata_hash] = Digest::MD5.hexdigest(wgm.e_workgroup.metadata_profile.to_hash.to_s)
      end

      if wgm.e_workgroup.dropbox
        m[:dbox_require_ear] = wgm.e_workgroup.dbox_require_ear
      end

      m
    }

    out
  end

  #add/remove e_memberships
  def update_memberships(saml_response, current_user_ad_groups=[])
    active_saml_response_ad_groups = (saml_response.attributes.multi(:member_of) || []).collect { |dn| AdGroup.active.find_by_dn(dn) }.compact
    obsolete_groups = current_user_ad_groups - active_saml_response_ad_groups
    new_groups = active_saml_response_ad_groups - current_user_ad_groups
    unaffected_group_memberships = active_saml_response_ad_groups.collect { |group| group.ad_group_memberships }
    obsolete_group_memberships = obsolete_groups.collect { |group| group.ad_group_memberships }
    obsolete_workgroups =
      obsolete_group_memberships.flatten.map(&:e_workgroup_id) - unaffected_group_memberships.flatten.map(&:e_workgroup_id)
    new_group_memberships = new_groups.collect { |group| group.ad_group_memberships }
    new_workgroups =
      new_group_memberships.flatten.map(&:e_workgroup_id) + unaffected_group_memberships.flatten.map(&:e_workgroup_id)

    (obsolete_workgroups - new_workgroups).uniq.each do |wg_id|
      membership = self.e_memberships.find_by_e_workgroup_id(wg_id)
      membership.destroy if membership
    end
    (new_workgroups - obsolete_workgroups).each do |wg_id|
      membership = self.e_memberships.find_or_create_by_e_workgroup_id(wg_id)
    end

    set_group_membership_types active_saml_response_ad_groups if active_saml_response_ad_groups.any?
  end

  def set_group_membership_types(active_ad_groups)
    most_permissive_membership_type = {}
    active_ad_group_memberships = active_ad_groups.collect { |group| group.ad_group_memberships }.flatten!

    ad_group_memberships_by_workgroup = active_ad_group_memberships.group_by { |membership| membership[:e_workgroup_id] }

    ad_group_memberships_by_workgroup.each do |e_workgroup_id, memberships|
      standard_access = memberships.detect do |membership|
        membership.access_level == AdGroupMembership::ACCESS_STANDARD
      end
      most_permissive_membership_type[e_workgroup_id] = standard_access ? AdGroupMembership::ACCESS_STANDARD : AdGroupMembership::ACCESS_SUBMIT_ONLY
    end

    active_ad_group_memberships.each do |ad_group_membership|
      membership = EMembership.find_by_user_id_and_e_workgroup_id(self.id, ad_group_membership.e_workgroup_id)
      membership.update_attributes(access_level: most_permissive_membership_type[ad_group_membership.e_workgroup_id]) unless membership.type_individually_set?
    end
  end

  ##################################################
  # Sender quota for the user
  ##################################################

  # Total data in MB sent during sender quota duration
  def completed_session_data
    sessions = ESession.find(:all, :conditions => ["created_at > ? and initiator_user_id = ? and status != ? and (direction = ? OR direction = ?)", EConfiguration.sender_quota_duration.to_i.hours.ago, self, "Transferring", 'Upload', 'Create'])
    total_bytes = 0
    sessions.each do |session|
      unless session.e_package.sender_quota_status == ESession::STATUS_QUOTA_REACHED
        if sender_email_whitelisted?(session.delivery.sender) || recipient_email_whitelisted?(session.delivery.recipient_emails.split(',')) || recipient_email_whitelisted?(session.delivery.wg_recipient_emails)
          next
        else
         total_bytes += session.bytes_total
        end
      end
    end

    total_bytes_in_mb = total_bytes / (1024.0 * 1024.0)

    return total_bytes_in_mb
  end

  # Total data in MB being transferred in the ongoing sessions
  def ongoing_session_data
    sessions = ESession.find(:all, :conditions => ["created_at > ? and initiator_user_id = ? and status = ? and (direction = ? OR direction = ?)", EConfiguration.sender_quota_duration.to_i.hours.ago, self, "Transferring", 'Upload', 'Create'])
    total_bytes = 0
    sessions.each do |session|
      total_bytes += session.bytes_total
    end

    total_bytes_in_mb = total_bytes / (1024.0 * 1024.0)

    return total_bytes_in_mb
  end

  # Returns true if the sender is whitelisted from sender quota limitation
  def sender_email_whitelisted?(sender = self)
    whitelist = EConfiguration.quota_sender_whitelist

    if whitelist && whitelist.select { |e| sender.email.downcase.end_with?(e) }.present?
      return true
    end

    return false
  end

  # Returns true if the recipient is whitelisted from sender quota limitation
  def recipient_email_whitelisted?(recipient_emails)
    return false if recipient_emails.nil? || recipient_emails.blank?

    whitelist = EConfiguration.quota_recipient_whitelist

    return false if !whitelist || whitelist.empty?
    recipient_emails.each do |recipent|
      if whitelist.select { |e| recipent.downcase.end_with?(e) }.any?
        return true
      else
        next
      end
    end

    return false
  end

  # returns true if the sender quota is enabled or set to default in the security settings
  def user_quota_enabled?

    if global_sender_quota_enabled?
      if self.override_security_sender_quota
        return false if !self.user_quota_limit
      end

      return true
    end

    return false
  end

  # Check if the sender quota is enabled at the global level security settings
  def global_sender_quota_enabled?
    if (EConfiguration.sender_quota_enabled? || EConfiguration.sender_quota_default?) && !sender_email_whitelisted?
      return true
    else
      return false
    end
  end

  # Total remaining data limit available in the sender quota duration window
  def quota_limit_available
    return user_quota_limit - completed_session_data
  end

  # Return sender quota limit for the current user
  def user_quota_limit
    if EConfiguration.sender_quota_default? && self.override_security_sender_quota
      return self.sender_quota_limit
    else
      return EConfiguration.first.sender_quota_limit
    end
  end

  private

  def sanitize_and_strip_tags
    self.name = Utils.sanitize_and_strip_tags(self.name) if self.name_changed? && !external_email_user? && !(saml? or ldap?)
    self.email = Utils.sanitize_and_strip_tags(self.email) if self.email_changed?  && !(saml? or ldap?)
    self.first_name = Utils.sanitize_and_strip_tags(self.first_name) if self.first_name_changed?  && !(saml? or ldap?)
    self.last_name = Utils.sanitize_and_strip_tags(self.last_name) if self.last_name_changed?  && !(saml? or ldap?)
    self.received_notification_cc = Utils.sanitize_and_strip_tags(self.received_notification_cc) if self.received_notification_cc_changed?
    self.description = Utils.sanitize_and_strip_tags(self.description) if self.description_changed?
  end

  # certain attributes are not fields in the database and need to be handled appropriately
  def set_special_whitelisted_attributes(whitelist, attribute_hash)
    if whitelist.include? :expiration_date
      self.expiration_date = attribute_hash[:expiration_date] || Utils.convert_date(attribute_hash, :expiration_date)
    end
    if whitelist.include? :account_activated
      record_activity
    end
    self.home_server_name = attribute_hash[:home_server_name] if (whitelist.include? :home_server_name) && attribute_hash[:home_server_name]
    self.password = attribute_hash[:password] if (whitelist.include? :password) && !attribute_hash[:password].nil?
    self.password_confirmation = attribute_hash[:password_confirmation] if (whitelist.include? :password_confirmation) && !attribute_hash[:password_confirmation].nil?
    self.delete_after_download_policy = attribute_hash[:delete_after_download_policy] if (whitelist.include? :delete_after_download_policy) && !attribute_hash[:delete_after_download_policy].nil?
    self.role = attribute_hash[:role] if (whitelist.include? :role) && !attribute_hash[:role].nil?
    self.account_activated = attribute_hash[:account_activated] if (whitelist.include? :account_activated) && !attribute_hash[:account_activated].nil?
  end

  def self.set_ad_user_attributes(user, attributes)
    whitelist = [
      :account_expires,
      :ad_individual_import,
      :ad_objectguid,
      :allow_receipt_cc_edit,
      :allow_user_specified_delete_after_download,
      :allowed_ip_addresses_for_download,
      :allowed_ip_addresses_for_login,
      :allowed_ip_addresses_for_upload,
      :kbps_initial_target_download_rate,
      :kbps_initial_target_upload_rate,
      :kbps_max_target_download_rate,
      :kbps_max_target_upload_rate,
      :lock_minrate_policy,
      :can_send_normal_packages,
      :can_send_to_all_faspex_users,
      :can_send_to_external_email,
      :delete_after_download_policy,
      :description,
      :downloads_allowed,
      :email,
      :expiration_date,
      :first_name,
      :forwarding_allowed,
      :can_create_from_remote,
      :home_server_id,
      :last_name,
      :ldap_attributes,
      :name,
      :notify_on_download,
      :notify_downloader_on_download,
      :notify_on_receive,
      :notify_on_validation_failure_if_sender,
      :notify_on_validation_failure_if_recipient,
      :override_home_server,
      :override_server_delete_after_download,
      :override_transfer_parameters,
      :password,
      :password_expiration_interval,
      :password_expires,
      :enforce_password_history,
      :password_history_limit,
      :received_notification_cc,
      :role,
      :uploads_allowed,
      :account_activated,
      :type,
      :can_invite,
      :can_have_public_url,
      :hide_user_dir,
      :show_only_completed_packages
    ]
    user.set_whitelisted_attributes(whitelist, attributes)
  end

  ##
  # Active Directory Scheduled Sync
  ##

  def self.last_ad_sync_operation
    conditions = ["affected_id = ? AND method = 'sync_ad_users_with_ad_server'", -1]
    AdBackgroundOperation.find(:first,
                               :order => "updated_at DESC, id DESC",
                               :conditions => conditions)
  end

  def self.needs_periodic_ad_sync?
    AdSynchronizer.sync_operation_is_older_than_periodic_threshold?(last_ad_sync_operation)
  end

  def self.schedule_periodic_ad_sync
    AdBackgroundOperation.initiateScheduledOperation("Sync All Users", "sync_ad_users_with_ad_server", {}, "", 1)
  end

  def login_allowed_by_slave?(remote_ip)
    login_from_ip_allowed?(remote_ip) && HomeServer.this_server == self.home_server
  end

  def authenticated_by_master?(password)
    Utils.log_debug("Relaying authentication request to master server")
    response = send_remote_login_request(password)
    Utils.log_debug("Response: #{response.class}, response body: #{response.body}")
    # answer can be "success", "success: user not on this home server" or "failure"
    (response.class == Net::HTTPOK && response.body.start_with?("success")) ? true : false
  rescue
    Utils.log_error($!)
    false
  end

  def send_remote_login_request(password)
    master = HomeServer.master_server
    master_path = master.url_namespace + authenticate_ds_user_login_path
    post_http_login_request(master.hostname, master.https_port, master_path, "username" => self.name, "password" => password)
  end

  def post_http_login_request(host, port, path, payload_data)
    https = Net::HTTP.new(host, port)
    https.use_ssl = MultiServer::Config.use_ssl
    https.verify_mode = OpenSSL::SSL::VERIFY_NONE
    https.read_timeout = MultiServer::Config.http_timeout
    request = Net::HTTP::Post.new(path)
    request.set_form_data(payload_data)
    request["Accept"] = "text/plain"
    # This block evaluates to a Net::HTTPResponse object, unless https.start raises an exception
    return https.start do |http|
      http.request(request)
    end
  end


  def ad_authenticate(ad_objectguid, password)
    user_objectguid = authorization_domain.directory_service.binary_uuid? ? Utils.decode_bytes_from_hex(ad_objectguid) : ad_objectguid
    user = LdapUser.first(authorization_domain, authorization_domain.directory_service.uuid_search_attribute => user_objectguid)
    user.authenticate(password)
  rescue => e
    logger.warn("DS authentication failed: #{e}\n#{e.backtrace.join("\n")}")
    false
  end

  # to avoid having an external user that can't be sent to
  def remove_internal_user_from_contacts
    Contact.destroy_all(["email = ?", name]) unless is_a?(ExternalUser)
  end

  end
