require_dependency 'transfer_cookie'

module UrlRegeneration
  class Regenerator
    class InvalidXML < StandardError; end

    def self.issue_token_for_urls(xml, requesting_user_id, requesting_ip, requested_host)
      begin
        urls = parse(xml)
        token = refresh(urls, requesting_user_id, requesting_ip, requested_host)
        return token
      rescue XmlParser::InvalidXML
        regeneration_failure(xml, "Cannot parse XML", "xml-not-well-formed")
      rescue URI::Error
        regeneration_failure(xml, "Could not parse URL: #{$!.to_s}", "invalid-url")
      rescue InvalidTransferCookie
        regeneration_failure(xml, "Url contains invalid cookie: #{$!.to_s}", "invalid-url")
      rescue ActiveRecord::RecordNotFound
        regeneration_failure(xml, $!.to_s, "url-not-found")
      rescue Errors::RegenerationError
        raise $!
      rescue Delivery::TransferSpecError
        regeneration_failure(xml, $!.to_s, "url-not-found")
      rescue SecurityError
        regeneration_failure(xml, $!.to_s, "unauthorized")
      rescue Exception
        Utils.log_info("Error regenerating URLs: #{$!.to_s}")
        Utils.log_debug("#{$!.backtrace.join("\n")}")
        regeneration_failure(xml, $!.to_s, "unknown-error")
      end
    end

    def self.regeneration_failure(xml, err, cond)
      raise Errors::RegenerationError.new(err.to_s, cond)
    end

    def self.parse(raw_xml)
      xml = XmlParser.parse_xml(raw_xml)
      unless xml.errors.blank?
        xml = convert_to_valid_XML(raw_xml) # ADDED TO ACCOUNT FOR FASPX-1544
      end
      url_list = xml/'url-list'
      validate_list_element(url_list)
      url_elems = url_list/'url'
      url_elems.collect{|url| parse_url_element(url)}
    end

    # @param raw_xml [String] the raw_xml in the request_body of the REST call
    # Converts all XML invalid ampersands of "&" to "&amp;" for FASPX-1544
    # There are only 5 XML escape characters that use ampersands otherwise, after taking into account all of these
    # cases, we perform a swap for all relevant "&"s in the raw XML.
    # @return [String] properly formatted XML
    def self.convert_to_valid_XML(raw_xml)
      raw_xml = raw_xml.gsub("&lt;","LESS_THAN_XML")
      raw_xml = raw_xml.gsub("&gt;","GREATER_THAN_XML")
      raw_xml = raw_xml.gsub("&quot;", "QUOTATION_MARK_XML")
      raw_xml = raw_xml.gsub("&apos;", "APOSTROPHE_XML")
      raw_xml = raw_xml.gsub("&amp;", "AMPSERSAND_XML")
      raw_xml = raw_xml.gsub("&","&amp;")
      raw_xml = raw_xml.gsub("LESS_THAN_XML","&lt;")
      raw_xml = raw_xml.gsub("GREATER_THAN_XML","&gt;")
      raw_xml = raw_xml.gsub("QUOTATION_MARK_XML", "&quot;")
      raw_xml = raw_xml.gsub("APOSTROPHE_XML", "&apos;")
      raw_xml = raw_xml.gsub("AMPSERSAND_XML", "&amp;")
      Nokogiri::XML(raw_xml)
    end

    def self.validate_list_element(url_list)
      raise Errors::RegenerationError.new("Expected at least one 'url-list' element",
                                          "xml-not-well-formed") if (url_list.length == 0)
      raise Errors::RegenerationError.new("Expected at most one 'url-list' element",
                                          "xml-not-well-formed") if (url_list.length > 1)
    end

    # Parses faspe URLs for REST calls like issue-token
    def self.parse_url_element(url)
      validate_url_element(url)
      href = url['href'].to_s
      unless href =~ %r{^(fasp|faspe)://([^/]*)@([^/]*):(\d*)(/.*)\?.*cookie=([^&]*).*tags64=([^&]*).*}
        raise Errors::RegenerationError.new("Invalid url, unable to parse fasp_url", "invalid_url")
      end
      user, host, port, path, cookie, tags = $2, $3, $4, $5, $6, $7
      params = TransferParams.new(
        remote_user: user,
        remote_host: host,
        ssh_port: port,
        cookie: cookie,
        tags: Base64.strict_decode64(tags)
      )
      FaspUrl.new(path, params)
    end

    def self.validate_url_element(url)
      raise Errors::RegenerationError.new("Invalid url, missing 'href' attribute: #{url.to_s}",
                                          "invalid-url") if url['href'].nil?
    end

    def self.parse_outer_package_directory_from_url(url)
      url_paths = URI.decode(url.path).split('/')
      # Get the outer package name -- (docroot/)outer_package_name.aspera-package/inner_package_name
      # Figure out which one by looking for (uuid).aspera.package
      url_paths.select{ |p| p =~ /.*[a-zA-Z0-9\-]+\.aspera\-package/}.first rescue nil
    end

    def self.refresh(urls, requesting_user_id, requesting_ip, requested_host)
      raise ArgumentError, "issue-token request must contain at least one fasp url" if ( urls.class != Array or urls.length == 0 )

      directory_name = parse_outer_package_directory_from_url(urls[0])

      # verify that all urls being refreshed belong to the same package and are for download
      urls.map { |url|
        raise Errors::RegenerationError.new("Unable to generate token for urls from different packages",
                                            "invalid-url") if parse_outer_package_directory_from_url(url) != directory_name
        cookie = TransferCookie.new(url.transfer_params.cookie)
        raise Errors::RegenerationError.new("Invalid url, url is not for download", "invalid-url") if !cookie.is_download?
      }

      # verify that the package still exists and the user is authorized to download it
      package = EPackage.find_by_outer_directory_name_and_e_package_contents_exist(directory_name, true)
      raise ActiveRecord::RecordNotFound, "URL references package that no longer " +
        "exists: '#{urls[0].to_s}'" if package.nil?
      user = User.find(requesting_user_id)
      raise ActiveRecord::RecordNotFound, "User not found" if user.nil?
      raise SecurityError, "User is not authorized to download this package" if !package.can_be_downloaded_by?(user, requesting_ip)

      return refresh_download(urls, package, requesting_user_id, requesting_ip, requested_host)
    end

    def self.refresh_download(urls, package, requesting_user_id, requesting_ip, requested_host)
      relative_paths = []
      urls.each do |url|
        path = URI.decode(url.path)
        relative_paths << path_relative_to_package_root(package, path)
      end
      requesting_user = User.find_by_id(requesting_user_id)
      # Add the relay share_id to the options if it's in the tag, otherwise don't
      options = {token: true}
      tags = urls.first.transfer_params.tags
      tags =~ %r{.*"share":([0-9]*).*}
      options[:share] = $1 unless $1.nil?
      spec = package.deliveries.first.get_transfer_spec(
        ESession::DIRECTION_DOWNLOAD, relative_paths, requesting_user, options
      )
      spec.token
    end

    private

    def self.path_relative_to_package_root(package, path)
      package_root = Pathname.new(package.inner_directory_path(true))
      Pathname.new(path).relative_path_from(package_root).to_s
    end
  end
end
